'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _stringify = require('babel-runtime/core-js/json/stringify');

var _stringify2 = _interopRequireDefault(_stringify);

var _babelCodeFrame = require('babel-code-frame');

var _babelCodeFrame2 = _interopRequireDefault(_babelCodeFrame);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function codeGen(t, ast, _ref) {
  var getMergeBqlQueries = _ref.getMergeBqlQueries,
      file = _ref.file;

  getMergeBqlQueries();
  function error(msg, node) {
    throw new Error(msg + '\n\n' + (0, _babelCodeFrame2.default)(file.code, node.loc.start.line, node.loc.start.column));
  }
  function assert(value, msg, node) {
    if (!value) {
      throw error(msg, node);
    }
  }
  var JSONStringify = t.memberExpression(t.identifier('JSON'), t.identifier('stringify'));
  function genFieldSet(ast) {
    assert(ast.type === 'FieldSet', 'Expected FieldSet', ast);
    var properties = ast.body.filter(function (node) {
      return node.type !== 'Spread' && node.type !== 'Conditional';
    }).map(genNode);
    var merge = ast.body.filter(function (node) {
      return node.type === 'Spread';
    });
    var conditionals = ast.body.filter(function (node) {
      return node.type === 'Conditional';
    });
    var obj = t.objectExpression(properties);
    obj._isBqlOutput = true;
    merge.forEach(function (node) {
      obj = t.callExpression(getMergeBqlQueries(), [obj, node.arg.val]);
      obj._isBqlOutput = true;
    });
    conditionals.forEach(function (_ref2) {
      var condition = _ref2.condition,
          consequent = _ref2.consequent,
          alternate = _ref2.alternate;

      obj = t.callExpression(getMergeBqlQueries(), [obj, t.conditionalExpression(condition.val, genFieldSet(consequent), alternate ? genFieldSet(alternate) : t.nullLiteral())]);
      obj._isBqlOutput = true;
    });
    return obj;
  }
  function isConstantValue(value) {
    return value.type === 'String' || value.type === 'Bool' || value.type === 'Number';
  }
  function genField(ast) {
    assert(ast.type === 'Field', 'Expected Field', ast);
    var name = ast.name;
    var value = null;
    if (ast.body) {
      value = genFieldSet(ast.body);
    } else {
      value = t.booleanLiteral(true);
    }
    if (!ast.args || ast.args.length === 0) {
      return t.objectProperty(ast.alias ? t.stringLiteral(name + ' as ' + ast.alias) : t.identifier(name), value);
    } else {
      var push = function push(str) {
        if (typeof parts[parts.length - 1] === 'string') {
          parts[parts.length - 1] += str;
        } else {
          parts.push(str);
        }
      };

      var args = ast.args.slice().sort(function (a, b) {
        return a.name < b.name ? -1 : 1;
      });
      var parts = [name + '('];

      args.forEach(function (arg, i) {
        assert(arg.type === 'Arg', 'Expected Arg', arg);
        if (i !== 0) {
          push(',');
        }
        push(arg.name);
        push(':');
        if (isConstantValue(arg.val)) {
          push((0, _stringify2.default)(arg.val.val));
        } else {
          parts.push(arg.val.val);
        }
      });
      push(')');
      if (ast.alias) {
        push(' as ' + ast.alias);
      }
      if (parts.length === 1) {
        return t.objectProperty(t.stringLiteral(parts[0]), value);
      } else {
        return t.objectProperty(parts.map(function (part) {
          return typeof part === 'string' ? t.stringLiteral(part) : t.callExpression(JSONStringify, [part]);
        }).reduce(function (left, right) {
          return t.binaryExpression('+', left, right);
        }), value, true // computed
        );
      }
    }
  }
  function genNode(ast) {
    switch (ast.type) {
      case 'FieldSet':
        return genFieldSet(ast);
      case 'Field':
        return genField(ast);
      default:
        throw error('Unexpeced node type ' + ast.type, ast);
    }
  }
  return genNode(ast);
}
exports.default = codeGen;