'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (babel) {
  var t = babel.types;

  function isBqlReference(node) {
    // TODO: do this better
    return t.isIdentifier(node, { name: 'bql' });
  }
  function isCallTo(node, name) {
    return t.isCallExpression(node) && t.isIdentifier(node.callee) && node.callee.name === name;
  }
  // check if a variable is every mutated
  var mutatedVisitor = {
    Identifier: function Identifier(path, state) {
      if (path.node.name === state.name) {
        var node = path.parentPath.node;
        var key = path.parentKey;
        var isMutation = t.isAssignmentExpression(node) && key === 'left' || t.isUpdateExpression(node);
        if (isMutation) {
          state.isMutated = true;
          path.stop();
        }
      }
    }
  };

  // check if an expression is "safeish" to hoist
  var immutabilityVisitor = {
    ReferencedIdentifier: function ReferencedIdentifier(path, state) {
      if (path.node.name === 'JSON') {
        return;
      }
      // the only identifiers we'll treat as immutable are mergeBqlQueries and the queries being merged
      if (!state.mergeBqlQueries) {
        state.isImmutable = false;
        path.stop();
        return;
      }
      if (path.node.name === state.mergeBqlQueries.name) {
        return;
      }
      var parent = path.parentPath;
      while (!t.isStatement(parent.node) && !isCallTo(parent.node, state.mergeBqlQueries.name)) {
        if (t.isObjectExpression(parent.node)) {
          state.isImmutable = false;
          path.stop();
          return;
        }
        parent = parent.parentPath;
      }
      var s = { isMutated: false, name: path.node.name };
      state.filePath.traverse(mutatedVisitor, s);
      if (s.isMutated) {
        state.isImmutable = false;
        path.stop();
      }
    },
    enter: function enter(path, state) {
      if (path.isImmutable()) {
        return;
      }
      if (path.isIdentifier()) {
        return;
      }
      if (state.mergeBqlQueries && isCallTo(path.node, state.mergeBqlQueries.name)) {
        return;
      }
      if (t.isCallExpression(path.node) && t.isMemberExpression(path.node.callee) && t.isIdentifier(path.node.callee.object) && path.node.callee.object.name === 'JSON' && path.node.callee.computed === false && t.isIdentifier(path.node.callee.property) && path.node.callee.property.name === 'stringify') {
        return;
      }

      // see if the node is constant providing its children are constant
      var isConstant = t.isArrayExpression(path.node) || t.isBinaryExpression(path.node) || t.isBooleanLiteral(path.node) || t.isConditionalExpression(path.node) || t.isLogicalExpression(path.node) || t.isMemberExpression(path.node) || t.isNullLiteral(path.node) || t.isNumericLiteral(path.node) || t.isObjectExpression(path.node) || t.isObjectMethod(path.node) || t.isObjectProperty(path.node) || t.isParenthesizedExpression(path.node) || t.isSequenceExpression(path.node) || t.isSpreadElement(path.node) || t.isSpreadProperty(path.node) || t.isStringLiteral(path.node) || t.isTemplateLiteral(path.node) || false;
      if (isConstant) {
        return;
      }
      state.isImmutable = false;
      path.stop();
    }
  };
  function _getMergeBqlQueries(file) {
    if (file._mergeBqlQueries) return file._mergeBqlQueries;
    file._mergeBqlQueries = file.scope.generateUidIdentifier('mergeBqlQueries');
    var helper = buildHelper({
      MERGE_QUERIES: file._mergeBqlQueries
    });
    helper.body._compact = true;
    helper._generated = true;
    file.path.unshiftContainer("body", helper);
    return file._mergeBqlQueries;
  }
  return {
    visitor: {
      TaggedTemplateExpression: function TaggedTemplateExpression(path) {
        var _this = this;

        if (isBqlReference(path.node.tag)) {
          var tokens = (0, _lexer2.default)(path.node.quasi, this.file);
          var ast = (0, _parser2.default)(tokens, this.file);
          var transformed = (0, _codeGen2.default)(t, ast, {
            file: this.file,
            getMergeBqlQueries: function getMergeBqlQueries() {
              return _getMergeBqlQueries(_this.file);
            }
          });
          path.replaceWith(transformed);
        }
      },
      ObjectExpression: function ObjectExpression(path) {
        if (!path.node._isBqlOutput || path.node._hoisted) {
          return;
        }

        var state = { isImmutable: true, mergeBqlQueries: this.file._mergeBqlQueries, filePath: this.file.path };
        path.traverse(immutabilityVisitor, state);

        if (state.isImmutable) {
          path.hoist();
        } else {
          path.node._hoisted = true;
        }
      },
      CallExpression: function CallExpression(path) {
        if (!path.node._isBqlOutput || path.node._hoisted) {
          return;
        }

        var state = { isImmutable: true, mergeBqlQueries: this.file._mergeBqlQueries, filePath: this.file.path };
        path.traverse(immutabilityVisitor, state);

        if (state.isImmutable) {
          path.hoist();
        } else {
          path.node._hoisted = true;
        }
      }
    }
  };
};

var _babelTemplate = require('babel-template');

var _babelTemplate2 = _interopRequireDefault(_babelTemplate);

var _lexer = require('./lexer');

var _lexer2 = _interopRequireDefault(_lexer);

var _parser = require('./parser');

var _parser2 = _interopRequireDefault(_parser);

var _codeGen = require('./code-gen');

var _codeGen2 = _interopRequireDefault(_codeGen);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var buildHelper = (0, _babelTemplate2.default)('\n  function MERGE_QUERIES(firstQuery, secondQuery) {\n    if (secondQuery === null) {\n      return firstQuery;\n    }\n    const result = {};\n    let firstQueryUpdated = false;\n    let secondQueryUpdated = false;\n    for (let key in firstQuery) {\n      if (key in firstQuery) {\n        if (!secondQuery[key]) {\n          firstQueryUpdated = true;\n          result[key] = firstQuery[key];\n        } else if (firstQuery[key] !== secondQuery[key]) {\n          result[key] = MERGE_QUERIES(firstQuery[key], secondQuery[key]);\n          if (result[key] !== secondQuery[key]) {\n            firstQueryUpdated = true;\n            secondQueryUpdated = true;\n          }\n        } else {\n          result[key] = firstQuery[key];\n        }\n      }\n    }\n    if (!firstQueryUpdated) {\n      return secondQuery;\n    }\n    for (let key in secondQuery) {\n      if ((key in secondQuery) && !result[key]) {\n        secondQueryUpdated = true;\n        result[key] = secondQuery[key];\n      }\n    }\n    return secondQueryUpdated ? result : firstQuery;\n  }\n');