'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _stringify = require('babel-runtime/core-js/json/stringify');

var _stringify2 = _interopRequireDefault(_stringify);

var _babelCodeFrame = require('babel-code-frame');

var _babelCodeFrame2 = _interopRequireDefault(_babelCodeFrame);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var BRACKETS = ['(', ')', '{', '}'];

function trimStart(str) {
  return str.replace(/^\s+/, '');
}
function lex(_ref, file) {
  var quasis = _ref.quasis,
      expressions = _ref.expressions;

  var tokens = [];
  quasis.forEach(function (quasi, i) {
    var str = quasi.value.raw;
    function trim() {
      while (str !== trimStart(str) || str[0] === ',') {
        str = trimStart(str);
        if (str[0] === ',') str = str.substr(1);
      }
    }
    function lexKeyword() {
      var match = /^(if|else|as)\b/.exec(str);
      if (match) {
        var start = getLocation();
        str = str.substr(match[0].length);
        return { type: 'Keyword', val: match[0], loc: { start: start, end: getLocation() } };
      }
      return null;
    }
    function lexString() {
      if (str[0] === '"' || str[0] === '\'') {
        var start = getLocation();
        var quote = str[0];
        var val = '';
        var escaped = false;
        var _i = 1;
        while (_i < str.length && (escaped || str[_i] !== quote)) {
          if (str[_i] === '\\') escaped = true;
          if (str[_i] === '"' && !escaped) val += '\\';
          val += str[_i++];
        }
        if (str[_i] === quote) {
          str = str.substr(_i + 1);
          return { type: 'String', val: JSON.parse('"' + val + '"'), loc: { start: start, end: getLocation() } };
        } else {
          throw new Error('Missing closing quote\n\n' + (0, _babelCodeFrame2.default)(file.code, start.line, start.column));
        }
      }
      return null;
    }
    function lexBool() {
      var match = /^(true|false)\b/.exec(str);
      if (match) {
        var start = getLocation();
        str = str.substr(match[0].length);
        return { type: 'Bool', val: match[0] === 'true', loc: { start: start, end: getLocation() } };
      }
      return null;
    }
    function lexNumber() {
      var match = /^\-?[\d\_]+(\.[\d\_]+)?/.exec(str);
      if (match) {
        var start = getLocation();
        str = str.substr(match[0].length);
        return { type: 'Number', val: parseInt(match[0].replace(/\_/g, ''), 10), loc: { start: start, end: getLocation() } };
      }
      return null;
    }
    function lexBracket() {
      if (BRACKETS.indexOf(str[0]) !== -1) {
        var start = getLocation();
        var match = str[0];
        str = str.substr(1);
        return { type: 'Bracket', val: match, loc: { start: start, end: getLocation() } };
      }
      return null;
    }
    function lexColon() {
      if (str[0] === ':') {
        var start = getLocation();
        str = str.substr(1);
        return { type: 'Colon', loc: { start: start, end: getLocation() } };
      }
      return null;
    }
    function lexSpread() {
      if (str[0] === '.' && str[1] === '.' && str[2] === '.') {
        var start = getLocation();
        str = str.substr(3);
        return { type: 'Spread', loc: { start: start, end: getLocation() } };
      }
      return null;
    }
    function lexIdentifier() {
      var match = /^[a-zA-Z][a-zA-Z0-9]*\b/.exec(str);
      if (match) {
        var start = getLocation();
        str = str.substr(match[0].length);
        return { type: 'Identifier', val: match[0], loc: { start: start, end: getLocation() } };
      }
      return null;
    }
    function lexComment() {
      if (str[0] === '/' && str[1] === '/') {
        str = str.substr(str.indexOf('\n'));
        return true;
      }
      return false;
    }
    function getLocation() {
      var _quasi$loc$start = quasi.loc.start,
          line = _quasi$loc$start.line,
          column = _quasi$loc$start.column;

      column++;
      for (var _i2 = 0; _i2 < quasi.value.raw.length - str.length; _i2++) {
        if (quasi.value.raw[_i2] === '\n') {
          line++;
          column = 1;
        } else {
          column++;
        }
      }
      return { line: line, column: column };
    }
    function fail() {
      var loc = getLocation();
      throw new Error('Unexpected character ' + (0, _stringify2.default)(str[0]) + '\n\n' + (0, _babelCodeFrame2.default)(file.code, loc.line, loc.column));
    }
    trim();
    while (str.length) {
      if (!lexComment()) {
        tokens.push(lexKeyword() || lexString() || lexBool() || lexNumber() || lexBracket() || lexColon() || lexSpread() || lexIdentifier() || fail());
      }
      trim();
    }
    if (expressions[i]) {
      tokens.push({ type: 'Expression', val: expressions[i], loc: expressions[i].loc });
    }
  });
  var endLoc = quasis[quasis.length - 1].loc.end;
  tokens.push({ type: 'End', loc: { start: endLoc, end: endLoc } });
  return tokens;
}
exports.default = lex;